/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.engine.source;

import java.io.ByteArrayInputStream;
import java.io.Serializable;
import java.security.Principal;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.foundation.pdi.converter.Converter;
import sync4j.foundation.pdi.parser.ICalendarParser;
import sync4j.foundation.pdi.parser.XMLEventParser;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.source.AbstractSyncSource;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.server.store.NotFoundException;
import sync4j.framework.tools.Base64;

import sync4j.exchange.items.calendar.model.Calendar;
import sync4j.exchange.items.calendar.manager.CalendarManager;
import sync4j.exchange.items.calendar.CalendarParseException;
import sync4j.exchange.DataAccessException;
import sync4j.exchange.util.SourceUtils;

/**
 * This class define <i>SyncSource</i>
 * between SyncServer and Exchange Server calendar items
 *
 * @author Fabio Maggi @ Funambol
 *
 * version $Id: ExchangeCalendarSyncSource.java,v 1.14 2005/05/26 14:29:05 fabius Exp $
 */
public class ExchangeCalendarSyncSource
extends ExchangeSyncSource
implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    private static final String EXCHANGE_HREF_EXTENSION = ".eml";

    // --------------------------------------------------------------- Private data

    private CalendarManager cm      = null  ;

    // --------------------------------------------------------------- Constructors

    public ExchangeCalendarSyncSource() {
    }

    // --------------------------------------------------------------- Public Methods

   /**
    * @see SyncSource
    */
    public SyncItem setSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

        Calendar calendar       = null  ;
        Calendar newCalendar    = null  ;

        String   href           = null  ;
        String   content        = null  ;
        String   itemKey        = null  ;
        String   username       = null  ;
        String   credentials    = null  ;

        boolean  isAddCalendar  = false ;

        try {

            username    = ((Sync4jPrincipal)principal).getUsername()           ;

            credentials = ((Sync4jPrincipal)principal).getEncodedCredentials() ;

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // search href [eg. 1234.eml] in exchange items
            //
            href = getHref(itemKey);

            if (href == null) {
                //
                // if href not found, new item is created
                // default href = this.CurrentTimesMillis()
                //
                href = String.valueOf(System.currentTimeMillis()) +
                       EXCHANGE_HREF_EXTENSION;
                isAddCalendar = true;
            }

            calendar = getCalendarFromSyncItem(syncItem);

            calendar.setHref(href);

            this.cm = getCalendarManager();


            newCalendar = this.cm.setCalendar (calendar    ,
                                               username    ,
                                               credentials ,
                                               exchangeFolder);

            if (isAddCalendar) {
                toChangedItems(newCalendar.getId(), ITEM_ADD);
            } else {
                toChangedItems(newCalendar.getId(), ITEM_UPDATE);
            }


            syncItem = getSyncItem(newCalendar);

            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Exchange SyncSource " +
                           sourceURI              +
                           " - set syncItem"      +
                           syncItem.getKey().getKeyAsString());
            }

            return syncItem;

        } catch (Exception e) {

            if (isAddCalendar) {
                toChangedItems(newCalendar.getId(), ITEM_ADD_ERROR);
            } else {
                toChangedItems(newCalendar.getId(), ITEM_UPDATE_ERROR);
            }

            throw new SyncSourceException( "Error setting the item " +
                                           syncItem + " " + e.getMessage(), e) ;
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] setSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList() ;
        ArrayList ret              = new ArrayList() ;

        for (int i=0, l = syncItems.length; i < l; ++i) {
            try {
                ret.add(setSyncItem(principal, syncItems[i]));
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException("Error setting the following items: " +
                                          syncItemsInError.toString()
                                         );
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource " +
                       sourceURI              +
                       " - set syncItems");
        }

        return (SyncItem[])ret.toArray(new SyncItem[] {});

    }

    /**
     * @see SyncSource
     */
    public SyncItem getSyncItemFromId(Principal principal,
                                      SyncItemKey syncItemKey)
    throws SyncSourceException{

        SyncItem[] syncItems =
            getSyncItemsFromIds(principal, new SyncItemKey[] {syncItemKey});

        if ((syncItems == null) || (syncItems.length == 0)) {
            return null;
        }

        return syncItems[0];

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromIds(Principal principal,
                                          SyncItemKey[] syncItemKeys)
    throws SyncSourceException {

        ArrayList  syncItems    = null ;
        Calendar   calendar     = null ;

        String     username     = null ;
        String     credentials  = null ;

        String     id           = null ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        syncItems = new ArrayList();

        int l = syncItemKeys.length;

        for (int i=0; ((syncItemKeys != null) && (i < l)); ++i) {

            id = syncItemKeys[i].getKeyAsString();

            try {
                this.cm = getCalendarManager();
                calendar = this.cm.getCalendarById(username    ,
                                                   credentials ,
                                                   id          ,
                                                   this.exchangeFolder);
            } catch (DataAccessException e) {

                Throwable previous = e.getCause();

                if (previous instanceof NotFoundException) {

                    Logger log = Sync4jLogger.getLogger("source");
                    if (log.isLoggable(Level.SEVERE)) {
                        log.severe
                            ("Calendar not found while reading Exchange database: " +
                              e.getMessage()                                        );
                    }

                    Logger.
                        getLogger(LOG_NAME).
                            throwing(getClass().
                                getName(), "readExchangeDatabase", e);

                } else{
                    throw new SyncSourceException("Error reading items", e);
                }
            }

            if (calendar != null) {
                calendar.setState(SyncItemState.NEW ) ;
                syncItems.add(getSyncItem(calendar)     ) ;
            }

        }

        return (SyncItem[]) syncItems.toArray(new SyncItem[syncItems.size()]);

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getNewSyncItemKeys(Principal principal,
                                            Timestamp since    )
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getNewSyncItems(Principal principal,
                                      Timestamp since    )
    throws SyncSourceException {

        String ids[] = null;

        ids = this.getNewItemIds();

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "      +
                       sourceURI                   +
                       " - getting new sync items" );
        }

        if(ids == null || !(ids.length > 0)) {
            return new SyncItem[0];
        }

        return filterSyncItems(principal, ids, SyncItemState.NEW);

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Principal principal,
                                                Timestamp since    )
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getDeletedSyncItems(Principal principal,
                                          Timestamp since    )
    throws SyncSourceException {

        String  [] ids       = null ;
        SyncItem[] syncItems = null ;

        ids = this.getDeleteItemIds();

        int l = ids.length;

        syncItems = new SyncItem[l];

        for (int i = 0; i < l; i++) {

            syncItems[i] = new SyncItemImpl(this, ids[i], SyncItemState.DELETED);
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "          +
                       sourceURI                       +
                       " - getting deleted sync items" );
        }

        return syncItems;

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getUpdatedSyncItems(Principal principal, Timestamp since)
    throws SyncSourceException {

        String ids[] = null;

        ids = this.getUpdateItemIds();

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "          +
                       sourceURI                       +
                       " - getting updated sync items" );
        }

        if(ids == null || !(ids.length > 0)) {
            return new SyncItem[0];
        }

        return filterSyncItems(principal, ids, SyncItemState.UPDATED);

    }

    public SyncItemKey[] getUpdatedSyncItemKeys(Principal principal,
                                                Timestamp since)
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public void removeSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

        Calendar calendar    = null  ;

        String   username    = null  ;
        String   credentials = null  ;

        String   id          = null  ;
        String   href        = null  ;

        id          = syncItem.getKey().getKeyAsString()                    ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        href = getHref(id);

        try {
            if (href != null) {
                calendar = new Calendar(id)            ;
                calendar.setHref(href)                 ;
                this.cm.removeCalendar(calendar    ,
                                       username    ,
                                       credentials ,
                                       exchangeFolder)      ;
                toChangedItems(calendar.getId(), ITEM_REMOVE);
            } else {
                toChangedItems(calendar.getId(), ITEM_REMOVE_ERROR);
            }
        } catch (DataAccessException e) {
           toChangedItems(calendar.getId(), ITEM_REMOVE_ERROR);
           throw new SyncSourceException("Error reading items", e);
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "   +
                       sourceURI                +
                       " - removing sync items" );
        }


    }

    /**
     * @see SyncSource
     */
    public void removeSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList();

        for (int i=0, l = syncItems.length ; i < l; ++i) {
            try {
                removeSyncItem(principal, syncItems[i]);
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }  // next i

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException( "Error deleting the following items: "
                                         + syncItemsInError.toString()
                                         );
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "   +
                       sourceURI                +
                       " - removing sync items" );
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getAllSyncItems(Principal principal)
    throws SyncSourceException {
        return filterSyncItems(principal, SyncItemState.NEW);
    }

    /**
     * The same as <i>getSyncItemFromId()</i>
     * @see SyncSource
     */
    public SyncItem getSyncItemFromTwin(Principal principal, SyncItem syncItem)
            throws SyncSourceException {

        String username = ((Sync4jPrincipal) principal).getUsername();
        String credentials = ((Sync4jPrincipal) principal).getEncodedCredentials();
        SyncItem syncTwin = null;

        try {
            this.cm = getCalendarManager();
            Calendar calendar = getCalendarFromSyncItem(syncItem);
            Calendar twin = this.cm.getCalendarTwin(calendar,
                    username,
                    credentials,
                    exchangeFolder);

            if (twin != null) {
                syncTwin = getSyncItem(twin);
            }

        } catch (DataAccessException ex) {
            throw new SyncSourceException("Error getting twin", ex);
        }

        return syncTwin;
    }


    /**
     * The same as <i>getSyncItemFromIds()</i>
     *
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromTwins(Principal principal,
                                            SyncItem[] syncItems) throws
            SyncSourceException {

        ArrayList items = new ArrayList();

        for (int i = 0, l = syncItems.length; i < l; ++i) {
            items.add(getSyncItemFromTwin(principal, syncItems[i]));
        }

        return (SyncItem[]) items.toArray(new SyncItem[items.size()]);

    }
    // --------------------------------------------------------- Private methods

    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, ids, state
     *
     * @param principal principal
     * @param ids     an array containing calendar id
     * @param state   calendar state
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal  principal ,
                                       String[]   ids       ,
                                       char       state)
    throws SyncSourceException {

       Calendar[] calendars   = null ;

       String     username    = null ;
       String     credentials = null ;

       username    = ((Sync4jPrincipal) principal).getUsername()           ;
       credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

       try {

           this.cm = getCalendarManager();


           calendars = this.cm.getCalendars (username    ,
                                             credentials ,
                                             ids         ,
                                             this.exchangeFolder) ;

        } catch (DataAccessException e) {
            throw new SyncSourceException("Error reading calendar: " +
                                          e.getMessage(), e);
        }

        return getSyncItems(calendars, state);

    }

    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, state
     *
     * @param principal principal
     * @param ids     an array containing calendar id
     * @param state   calendar state
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal  principal ,
                                       char       state)
    throws SyncSourceException {

       Calendar[] calendars    = null ;

       String     username     = null ;
       String     credentials  = null ;

       username    = ((Sync4jPrincipal) principal).getUsername()           ;
       credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

       try {

           this.cm = getCalendarManager();

           calendars = this.cm.getAllCalendars (username    ,
                                                credentials ,
                                                this.exchangeFolder) ;

        } catch (DataAccessException e) {
            throw new SyncSourceException("Error reading calendar: " +
                                           e.getMessage(), e);
        }

        return getSyncItems(calendars, state);

    }

    /**
     * Create SyncItem array from Calendar array
     *
     * @param calendars
     * @param state
     * @return SyncItem array
     * @trows SyncSourceException
     */
    private SyncItem[] getSyncItems(Calendar[] calendars,
                                    char       state)
    throws SyncSourceException {

        SyncItem[] syncItems = null;

        int l = calendars.length;

        syncItems = new SyncItem [l];

        for (int i=0; i < l; i++) {
            calendars [i].setState(state)             ;
            syncItems [i] = getSyncItem(calendars[i]) ;
        }

        return syncItems;

    }

    /**
     * Create SyncItem from Calendar
     *
     * @param calendar
     * @return SyncItem
     * @trows SyncSourceException
     */
    private SyncItem getSyncItem(Calendar calendar)
    throws SyncSourceException {

        SyncItem syncItem       = null ;

        String   calendarStream = null ;

        syncItem = new SyncItemImpl(this                 ,
                                    calendar.getId    () ,
                                    calendar.getState ());

        try {

            calendarStream = calendar.toPP();

            if (TYPE_ICAL.equals(getType())) {
                calendarStream = getICalFromXml(calendarStream);
            }

            if (this.isEncode()) {
                syncItem.setProperty(
                    new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                     Base64.encode(calendarStream.getBytes())
                ));
            } else {
                syncItem.setProperty(
                    new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                     calendarStream.getBytes())
                );
            }

        } catch (CalendarParseException e) {
            throw new SyncSourceException( "Error setting the item "
                                          + syncItem
                                          , e
                                          );
        }

        return syncItem;

    }

    /**
     *
     * Create CalendarManager istance
     * (if not already exist)
     *
     * @trows DataAccessException
     */
    private CalendarManager getCalendarManager()
    throws DataAccessException {

        if (this.cm == null) {
            this.cm = new CalendarManager(this.getHost() ,
                                          this.getPort());
        }

        return this.cm;

    }


    /**
     *
     * Get XML message from ICal message
     * @param content message in ICal
     * @return message in XML
     *
     * @throws SyncSourceException
     */
    private String getXmlFromICal (String content)
    throws SyncSourceException {

        ByteArrayInputStream  buffer     = null ;
        Converter             converter  = null ;

        ICalendarParser       parser     = null ;

        sync4j.foundation.pdi.event.Calendar
                              calendar   = null ;

        String                xmlStream  = null ;

        Logger                log        = null ;

        try {

            calendar = new sync4j.foundation.pdi.event.Calendar();

            content = SourceUtils.handleLineDelimiting(content);

            buffer = new ByteArrayInputStream (content.getBytes());
            //converting the i-cal item into a Calendar object

            if ((content.getBytes()).length > 0) {
                log = Sync4jLogger.getLogger("source");

                parser = new ICalendarParser(buffer);
                parser.setLogger(log);
                calendar =
                    (sync4j.foundation.pdi.event.Calendar) parser.iCalendar();
            }

            //converting the Calendar object into an XML file
            converter = new Converter();
            xmlStream = converter.calendarToXML(calendar, tmz, convertDates);

        } catch (Exception e) {
            throw new SyncSourceException (e);
        }

        return xmlStream;

    }


    /**
     *
     * Get ICal message from XML message
     * @param content message in XML
     * @return message in ICal
     *
     * @throws SyncSourceException
     */
    private String getICalFromXml (String content)
    throws SyncSourceException {

        XMLEventParser              parser    = null ;
        Converter                   converter = null ;

        sync4j.foundation.pdi.event.Calendar
                                    calendar  = null ;

        String                      ical      = null ;


        try {

            parser = new XMLEventParser(new ByteArrayInputStream (content.getBytes()));
            calendar = (sync4j.foundation.pdi.event.Calendar) parser.
                        parse();
            //converting the Calendar object into a ical string
            converter = new Converter();
            ical = converter.calendarToIcalendar(calendar, tmz, convertDates);

        } catch (Exception e) {
            throw new SyncSourceException (e.getMessage());
        }

        return ical;
    }

    /**
     * Creates a new Calendar from the given syncItem
     * @param syncItem SyncItem
     * @return Calendar
     */
    private Calendar getCalendarFromSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        String content    = null;
        Calendar calendar = null;

        String itemKey = syncItem.getKey().getKeyAsString();

        byte[] itemContent =
                (byte[]) syncItem.getPropertyValue(
                        SyncItem.PROPERTY_BINARY_CONTENT);

        if (itemContent == null) {
            itemContent = new byte[0];
        }

        content = new String(itemContent);

        if (this.isEncode() && content != null && content.length() > 0) {
            content = new String(Base64.decode(content));
        }

        if (TYPE_ICAL.equals(getType())) {
            content = getXmlFromICal(content);
        }

        Date t =
                new Date(((Timestamp) syncItem.getPropertyValue(
                        SyncItem.PROPERTY_TIMESTAMP)).getTime());

        try {
            calendar = new Calendar(itemKey, content, t);
        } catch (CalendarParseException e) {
            throw new SyncSourceException("Error creating calendar", e);
        }

        return calendar;
    }
}
