/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.engine.source;

import java.io.Serializable;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.TimeZone;

import sync4j.framework.core.AlertCode;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.source.AbstractSyncSource;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.server.store.PersistentStore;
import sync4j.framework.server.store.PersistentStoreException;

import sync4j.exchange.items.common.model.Item;
import sync4j.exchange.items.common.manager.ItemManager;
import sync4j.exchange.DataAccessException;

import sync4j.framework.security.Sync4jPrincipal;

import sync4j.server.config.Configuration;

/**
 *
 * This class implements common methods of
 * <pre>
 *  ExchangeCalendarSyncSource
 *  ExchangeContactSyncSource
 *  ExchangeNoteSyncSource
 *  ExchangeTaskSyncSource
 * </pre>
 *
 * @author Fabio Maggi @ Funambol
 *
 * version $Id: ExchangeSyncSource.java,v 1.12 2005/05/26 14:30:22 fabius Exp $
 */
public abstract class ExchangeSyncSource
extends AbstractSyncSource
implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    protected static final String ITEM_ADD          = "A"                ;
    protected static final String ITEM_REMOVE       = "D"                ;
    protected static final String ITEM_UPDATE       = "U"                ;

    protected static final String ITEM_ADD_ERROR    = "AE"               ;
    protected static final String ITEM_REMOVE_ERROR = "DE"               ;
    protected static final String ITEM_UPDATE_ERROR = "UE"               ;

    protected static final String TYPE_VCARD        = "text/x-vcard"     ;
    protected static final String TYPE_ICAL         = "text/x-vcalendar" ;

    // -------------------------------------------------------------- Private data

    private   Item[]   localItems       = null  ;
    private   Item[]   exchangeItems    = null  ;

    private   String[] newItemIds       = null  ;
    private   String[] deleteItemIds    = null  ;
    private   String[] updateItemIds    = null  ;

    private   HashMap  changedItems     = null  ;

    protected TimeZone tmz              = null  ;

    protected boolean  convertDates     = false ;

    protected Logger   log              = null  ;

    // -------------------------------------------------------------- Properties

     /**
      * Property exchange folder
      */
     protected String exchangeFolder;

     /** Getter for property exchange folder.
      * @return Value of property exchange folder.
      *
      */
     public String getExchangeFolder() {
         return exchangeFolder;
     }

     /** Setter for property exchange folder.
      * @param exchangeFolder New value of property exchange folder.
      *
      */
     public void setExchangeFolder(String exchangeFolder) {
         this.exchangeFolder = exchangeFolder;
     }

    /**
     * Property ExchangeServer host
     */
    private String host;

    /** Getter for property exchangeServerHost.
     * @return Value of property exchangeServerHost.
     *
     */
    public String getHost() {
        return host;
    }

    /** Setter for property exchangeServerHost.
     * @param host New value of property exchangeServerHost.
     *
     */
    public void setHost(String host) {
        this.host = host;
    }

    /**
     * Property ExchangeServer port
     */
    private int port = 80;

    /** Getter for property exchangeServerPort.
     * @return Value of property exchangeServerPort.
     *
     */
    public int getPort() {
        return port;
    }

    /** Setter for property exchangeServerPort.
     * @param port New value of property exchangeServerPort.
     *
     */
    public void setPort(int port) {
        this.port = port;
    }

    /**
     * Should the content be encoded?
     */
    private boolean encode = false;

    /** Getter for property encode.
     * @return Value of property encode.
     *
     */
    public boolean isEncode() {
        return encode;
    }

    /** Setter for property encode.
     * @param encode New value of property encode.
     *
     */
    public void setEncode(boolean encode) {
        this.encode = encode;
    }

    /**
     * Property schedule
     */
    private String schedule;

    /** Getter for property schedule.
     * @return Value of property schedule.
     *
     */
    public String getSchedule() {
        return schedule;
    }

    /** Setter for property schedule.
     * @param schedule New value of property schedule.
     *
     */
    public void setSchedule(String schedule) {
        this.schedule = schedule;
    }

    /**
     * Property scheduleEnabled
     */
    private boolean scheduleEnabled;

    /** Getter for property scheduleEnabled.
     * @return Value of property scheduleEnabled.
     *
     */
    public boolean isSchedulEnablede() {
        return scheduleEnabled;
    }

    /** Setter for property schedule.
     * @param scheduleEnabled New value of property schedule.
     *
     */
    public void setScheduleEnabled(boolean scheduleEnabled) {
        this.scheduleEnabled = scheduleEnabled;
    }

    /** Getter for property newItemIds.
     * @return Value of property newItemIds.
     *
     */
    public String[] getNewItemIds()    {
        return this.newItemIds;
    }

    /** Getter for property updateItemIds.
     * @return Value of property updateItemIds.
     *
     */
    public String[] getUpdateItemIds() {
        return this.updateItemIds;
    }

    /** Getter for property deleteItemIds.
     * @return Value of property deleteItemIds.
     *
     */
    public String[] getDeleteItemIds() {
        return this.deleteItemIds;
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * @see SyncSource
     */
    public void beginSync(Principal principal, int syncMode)
    throws SyncSourceException {

        String       principalId    = null ;
        String       username       = null ;
        String       credentials    = null ;

        ItemManager  im             = null ;

        log = Sync4jLogger.getLogger("source");

        changedItems = new HashMap();

        principalId = ((Sync4jPrincipal) principal).getId()                 ;
        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        try {
            String deviceId      = ((Sync4jPrincipal)principal).getDeviceId();
            Sync4jDevice device  = getDevice             (deviceId)  ;
            String timezone      = device.getTimeZone    (        )  ;
            if (timezone != null && timezone.length() > 0)  {
                tmz                  = TimeZone.getTimeZone  (timezone)  ;
            }
            convertDates         = device.getConvertDate (        )  ;
        } catch (Exception e) {
            throw new SyncSourceException(e.getMessage());
        }


        try {

            im  = new ItemManager (this.host ,
                                   this.port);

            if ((syncMode != AlertCode.SLOW) &&
                (syncMode != AlertCode.REFRESH_FROM_SERVER)) {

                this.exchangeItems  = im.getExchangeItems  (username    ,
                                                            credentials ,
                                                            this.exchangeFolder ) ;

                this.localItems = im.getLocalItems (this.sourceURI,
                                                    principalId)  ;

                this.newItemIds     = findNewItemIds    () ;
                this.updateItemIds  = findUpdateItemIds () ;
                this.deleteItemIds  = findDeleteItemIds () ;

            }

            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Exchange SyncSource " + sourceURI + " - begineSync");
            }

        } catch (DataAccessException e) {
            throw new SyncSourceException("Error accessing items: " +
                                           e.getMessage(), e);
        }


    }

    /**
     * @see SyncSource
     */
    public void endSync(Principal principal)
    throws SyncSourceException {

        ItemManager  im                  = null  ;

        Item  []     exchItems           = null  ;
        String       id                  = null  ;

        String       principalId         = null ;
        String       username            = null ;
        String       credentials         = null ;

        principalId = ((Sync4jPrincipal) principal).getId()                 ;
        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        try {

            im  = new ItemManager (this.host ,
                                   this.port );

            //
            // to do: optimize the performance,
            //        by read only update / add items
            //        to setting lastModified
            //        with addItems[exist], deleteItems[exist],
            //        updateItems[implements] array
            //
            exchItems  = im.getExchangeItems  (username            ,
                                               credentials         ,
                                               this.exchangeFolder ) ;

            im.updateLocalItems (exchItems      ,
                                 this.sourceURI ,
                                 principalId);

            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Exchange SyncSource " + sourceURI + " - endSync");
                log.severe("Items changed: " + changedItems.size());

                Iterator i = changedItems.keySet().iterator();

                while(i.hasNext()) {

                    id = (String) i.next();

                    log.severe("Item id: "  + id +
                               " Status: "  + changedItems.get(id));
                }

            }

        } catch (DataAccessException e) {
            throw new SyncSourceException("Error updating items:" +
                                          e.getMessage(), e);
        }

    }

    // ---------------------------------------------------------- Private Methods

    /**
     * Create new items id array
     *
     * @return new items id array
     */
    private String[] findNewItemIds () {

        ArrayList  newRows       = null ;
        String[]   ids           = null ;
        Item       exchangeItem  = null ;

        int        count         = 0    ;

        newRows = new ArrayList();

        for (int i = 0, l = this.exchangeItems.length; i < l; i++) {

            exchangeItem = this.exchangeItems[i];

            if (!this.isItemInArray(this.localItems, exchangeItem)) {
                newRows.add(exchangeItem.getId());
                count++;
            }

        }

        return (String[])newRows.toArray(new String[count]) ;

    }

    /**
     * Create deleted items id array
     *
     * @return deleted items id array
     */
    private String[] findDeleteItemIds () {

        ArrayList  deleteRows       = null ;
        String[]   ids              = null ;
        Item       localItem        = null ;

        int        count            = 0;

        deleteRows = new ArrayList();

        for (int i=0, l = this.localItems.length; i < l; i++) {

           localItem = localItems[i];

           if (!this.isItemInArray(this.exchangeItems, localItem)) {
               deleteRows.add(localItem.getId());
               count++;
           }

        }

        return (String[])deleteRows.toArray(new String[count]) ;

    }

    /**
     * Create updated items id array
     *
     * @return updated items id array
     */
    private String[] findUpdateItemIds () {

        ArrayList  updateRows               = null ;
        String[]   ids                      = null ;

        Item       exchangeItem             = null ;
        Item       localItem                = null ;

        Date       excItemLastModified      = null ;
        Date       localItemLastModified    = null ;

        String     excItemLastModifiedStr   = null;
        String     localItemLastModifiedStr = null ;

        int        count                    = 0    ;

        updateRows = new ArrayList();

        for (int i = 0, l = this.exchangeItems.length; i < l; i++) {

            exchangeItem   = this.exchangeItems[i]                  ;
            localItem      = this.getItemById(this.localItems,
                                              exchangeItem.getId()) ;

            if (localItem != null) {

                 excItemLastModified      = exchangeItem.getLastModified () ;
                 localItemLastModified    = localItem.getLastModified    () ;

                 //
                 // because java.util.Date no save millisecond
                 // string rappresention of date is in HH:mm:ss
                 //
                 excItemLastModifiedStr   = String.valueOf(excItemLastModified)   ;
                 localItemLastModifiedStr = String.valueOf(localItemLastModified) ;

                 if (!excItemLastModifiedStr.equals(localItemLastModifiedStr)) {

                    updateRows.add(exchangeItem.getId());
                    count++;

                 }

             }

        }

        return (String[])updateRows.toArray(new String[count]) ;

    }

    /**
     * Return href (webdav resource name) of item
     *
     * @param id id of item to find
     * @return href (webdav resource name) of find item,
     *         or <p>null</p> if item not found
     */
    protected String getHref(String id) {

        Item    exchangeItem  = null ;
        String  href          = null ;

        //
        // in case of slow sync no exchange item load
        //
        if (exchangeItems != null) {

            for (int i=0, l = exchangeItems.length; i < l; i++) {

                exchangeItem = exchangeItems[i];

                if (exchangeItem.getId().equals(id)) {

                    href = exchangeItem.getHref();
                    break;
                }

             }

         }

         return href;

    }


    /**
     * Add item to mapp operations status array
     *
     * @param id id of item to add
     * @param status status of item to add
     *
     */
    protected void toChangedItems (String id, String status) {
        changedItems.put(id, status);
    }

    // ---------------------------------------------------------- Private Methods

    /**
     * get Item by id
     *
     * @param items array
     * @param id
     * @return <p>item</p> if find
     *         <p>null</p> if not find
     **/
    private Item getItemById(Item[] items,
                             String id) {

        Item item          = null;
        Item sequenceItem  = null;

        for (int i = 0, l = items.length; i < l; i++) {

            sequenceItem = items[i];

            if (sequenceItem.getId().equals(id)) {
                item = sequenceItem;
                break;
            }

        }

        return item;

    }

    /**
     *
     * check if item id is included
     * in ids from array of items
     *
     * @param items items array
     * @param id id to find
     * @return <p>true</p> if is included
     *         <p>false</p> if is not included
     **/
    private boolean isIdInArray(Item[] items,
                                String id) {

        boolean isFind = false;

        for (int i = 0, l = items.length; i < l; i++) {

            if (items[i].getId().equals(id)) {
                isFind = true;
                break;
            }

        }

        return isFind;

    }

    /**
     *
     * check if item is included
     * in array of items
     *
     * @param items items array
     * @param item item to find
     * @return <p>true</p> if is included
     *         <p>false</p> if is not included
     **/
    private boolean isItemInArray(Item[] items,
                                  Item   item) {

        return isIdInArray(items, item.getId());

    }


    /**
     * Return the device with the given deviceId
     * @param deviceId String
     * @return Sync4jDevice
     * @throws PersistentStoreException
     */
    private Sync4jDevice getDevice(String deviceId) throws PersistentStoreException {
        Sync4jDevice device = new Sync4jDevice(deviceId);
        PersistentStore store = Configuration.getConfiguration().getStore();
        store.read(device);
        return device;
    }


}
