/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.engine.source;

import java.io.Serializable;
import java.security.Principal;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.source.AbstractSyncSource;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.server.store.NotFoundException;
import sync4j.framework.tools.Base64;

import sync4j.exchange.items.task.model.Task;
import sync4j.exchange.items.task.manager.TaskManager;
import sync4j.exchange.items.task.TaskParseException;
import sync4j.exchange.DataAccessException;

/**
 * This class define <i>SyncSource</i>
 * between SyncServer and Exchange Server tasks items
 *
 * @author Fabio Maggi @ Funambol
 *
 * version $Id: ExchangeTaskSyncSource.java,v 1.8 2005/01/28 15:54:44 fabius Exp $
 */
public class ExchangeTaskSyncSource
extends ExchangeSyncSource
implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    private static final String EXCHANGE_HREF_EXTENSION = ".eml" ;

    // -------------------------------------------------------------- Private data

    private TaskManager tm = null;

    // ------------------------------------------------------------ Constructors

    public ExchangeTaskSyncSource() {
    }

    // ---------------------------------------------------------- Public Methods

   /**
    * @see SyncSource
    */
    public SyncItem setSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

       Task   task          = null  ;
       Task   newTask       = null  ;

       String itemKey       = null  ;
       String href          = null  ;
       String content       = null  ;

       String username      = null  ;
       String credentials   = null  ;

       boolean isAddTask    = false ;

       try {

            username    = ((Sync4jPrincipal)principal).getUsername()           ;
            credentials = ((Sync4jPrincipal)principal).getEncodedCredentials() ;

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // search href [eg. 1234.eml] in exchange items
            //
            href = getHref(itemKey);

            if (href == null) {
               //
               // if href not found, new item is created
               // default href = this.CurrentTimesMillis()
               //
               href = String.valueOf(System.currentTimeMillis()) +
                      EXCHANGE_HREF_EXTENSION;
               isAddTask = true;
            }


            byte[] itemContent =
               (byte[])syncItem.getPropertyValue(
                   SyncItem.PROPERTY_BINARY_CONTENT);

            if (itemContent == null) {
               itemContent = new byte[0];
            }

            content = new String (itemContent);

            if (this.isEncode() && content != null && content.length() > 0) {
                content = new String (Base64.decode(content));
            }

            Date t =
               new Date(((Timestamp) syncItem.getPropertyValue(
                   SyncItem.PROPERTY_TIMESTAMP)).getTime());

            task = new Task(itemKey, content, t);

            task.setHref(href);

            this.tm = getTaskManager();

            newTask = this.tm.setTask (task           ,
                                       username       ,
                                       credentials    ,
                                       exchangeFolder );

            if (isAddTask) {
                toChangedItems(newTask.getId(), ITEM_ADD);
            } else {
                toChangedItems(newTask.getId(), ITEM_UPDATE);
            }

            syncItem = getSyncItem(newTask);

            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Exchange SyncSource " +
                           sourceURI              +
                           " - set syncItem"      +
                           syncItem.getKey().getKeyAsString());
            }

            return syncItem;

        } catch (Exception e) {

            if (isAddTask) {
                toChangedItems(newTask.getId(), ITEM_ADD_ERROR);
            } else {
                toChangedItems(newTask.getId(), ITEM_UPDATE_ERROR);
            }

            throw new SyncSourceException( "Error setting the item " +
                                          syncItem + " " + e.getMessage(), e);
        }
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] setSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList();
        ArrayList ret              = new ArrayList();

        for (int i=0, l = syncItems.length; i < l; ++i) {
            try {
                ret.add(setSyncItem(principal, syncItems[i]));
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }  // next i

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException("Error setting the following items: " +
                                          syncItemsInError.toString()
                                         );
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource " +
                       sourceURI              +
                       " - set syncItems");
        }

        return (SyncItem[])ret.toArray(new SyncItem[] {});

    }

    /**
     * @see SyncSource
     */
    public SyncItem getSyncItemFromId(Principal principal,
                                      SyncItemKey syncItemKey)
    throws SyncSourceException{

        SyncItem[] syncItems =
            getSyncItemsFromIds(principal, new SyncItemKey[] {syncItemKey});

        if ((syncItems == null) || (syncItems.length == 0)) {
            return null;
        }

        return syncItems[0];

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromIds(Principal     principal,
                                          SyncItemKey[] syncItemKeys)
    throws SyncSourceException {

        ArrayList  syncItems   = null ;
        Task       task        = null ;

        String     username    = null ;
        String     credentials = null ;

        String     id          = null ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        syncItems = new ArrayList();

        int l = syncItemKeys.length;

        for (int i = 0; ((syncItemKeys != null) && (i < l)); ++i) {

            id = syncItemKeys[i].getKeyAsString();

            try {
                this.tm  = getTaskManager();
                task     = this.tm.getTaskById(username           ,
                                               credentials        ,
                                               id                 ,
                                               this.exchangeFolder);
            } catch (DataAccessException e) {
                Throwable previous = e.getCause();

                if (previous instanceof NotFoundException) {

                    Logger log = Sync4jLogger.getLogger("source");
                    if (log.isLoggable(Level.SEVERE)) {
                        log.severe("Task not found while reading Exchange database: " + e.getMessage());
                    }
                    Logger.getLogger(LOG_NAME).throwing(getClass().getName(), "readExchangeDatabase", e);
                } else{
                    throw new SyncSourceException("Error reading items", e);
                }
            }

            if (task != null) {
                task.setState (SyncItemState.NEW );
                syncItems.add (getSyncItem (task)    );
            }

        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource " +
                       sourceURI              +
                       " - getting syncItems" );
        }

        return (SyncItem[]) syncItems.toArray(new SyncItem[syncItems.size()]);

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getNewSyncItemKeys(Principal principal,
                                            Timestamp since    )
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getNewSyncItems(Principal principal,
                                      Timestamp since    )
    throws SyncSourceException {

        String ids[] = null;

        ids = this.getNewItemIds();

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "     +
                       sourceURI                  +
                       " - getting new syncItems" );
        }

        if(ids == null || !(ids.length > 0)) {
            return new SyncItem[0];
        }

        return filterSyncItems(principal, ids, SyncItemState.NEW);

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Principal principal,
                                                Timestamp since    )
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getDeletedSyncItems(Principal principal,
                                          Timestamp since    )
    throws SyncSourceException {

        String  [] ids       = null ;
        SyncItem[] syncItems = null ;

        ids = this.getDeleteItemIds();

        int l = ids.length;

        syncItems = new SyncItem[l];

        for (int i = 0; i < l; i++) {
            syncItems[i] = new SyncItemImpl(this, ids[i], SyncItemState.DELETED);
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "        +
                       sourceURI                     +
                       " - getting delete syncItems" );
        }

        return syncItems;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getUpdatedSyncItems(Principal principal, Timestamp since)
    throws SyncSourceException {

        String ids[] = null;

        ids = this.getUpdateItemIds();

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "        +
                       sourceURI                     +
                       " - getting update syncItems" );
        }

        if(ids == null || !(ids.length > 0)) {
            return new SyncItem[0];
        }

        return filterSyncItems(principal, ids, SyncItemState.UPDATED);
    }


    public SyncItemKey[] getUpdatedSyncItemKeys(Principal principal,
                                                Timestamp since)
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public void removeSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

        Task    task         = null  ;

        String  username     = null  ;
        String  credentials  = null  ;

        String  id           = null  ;
        String  href         = null  ;

        id = syncItem.getKey().getKeyAsString();

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        href = getHref(id);

        try {
            if (href != null) {
                task = new Task(id)        ;
                task.setHref(href)         ;
                this.tm = getTaskManager() ;
                this.tm.removeTask(task           ,
                                   username       ,
                                   credentials    ,
                                   exchangeFolder ) ;
                toChangedItems(task.getId(), ITEM_REMOVE) ;
            } else {
                toChangedItems(task.getId(), ITEM_REMOVE_ERROR) ;
            }
        } catch (DataAccessException e) {
           toChangedItems(task.getId(), ITEM_REMOVE_ERROR) ;
           throw new SyncSourceException("Error reading items", e);
        }
    }

    /**
     * @see SyncSource
     */
    public void removeSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList();

        for (int i=0, l = syncItems.length ; i < l; ++i) {
            try {
                removeSyncItem(principal, syncItems[i]);
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }  // next i

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException( "Error deleting the following items: "
                                         + syncItemsInError.toString()
                                         );
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "  +
                       sourceURI               +
                       " - removing syncItems" );
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getAllSyncItems(Principal principal)
    throws SyncSourceException {

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "     +
                       sourceURI                  +
                       " - getting all syncItems" );
        }

        return filterSyncItems(principal, SyncItemState.NEW);

    }

    /**
     * The same as <i>getSyncItemFromId()</i>
     * @see SyncSource
     */
    public SyncItem getSyncItemFromTwin(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

        return getSyncItemFromId(principal, syncItem.getKey());

    }

    /**
     * The same as <i>getSyncItemFromIds()</i>
     *
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromTwins(Principal principal,
                                            SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList items = new ArrayList();
        for (int i = 0, l = syncItems.length; i < l; ++i){
            items.add(getSyncItemFromId(principal, syncItems[i].getKey()));
        }

        return (SyncItem[])items.toArray(new SyncItem[items.size()]);

    }

    // --------------------------------------------------------- Private methods

    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, ids, state
     *
     * @param principal principal
     * @param ids     an array containing task id
     * @param state   note state
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal  principal ,
                                       String[]   ids       ,
                                       char       state)
    throws SyncSourceException {

        Task[] tasks       = null ;

        String username    = null ;
        String credentials = null ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        try {

            this.tm = getTaskManager();

            tasks = this.tm.getTasks     (username            ,
                                          credentials         ,
                                          ids                 ,
                                          this.exchangeFolder ) ;

        } catch (DataAccessException e) {
            throw new SyncSourceException("Error reading tasks: " +
                                         e.getMessage(), e);
        }

        return getSyncItems(tasks, state);

    }

    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal,state
     *
     * @param principal principal
     * @param ids     an array containing task id
     * @param state   note state
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal  principal ,
                                       char       state)
    throws SyncSourceException {

        Task[]     tasks     = null ;

        String username    = null ;
        String credentials = null ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        try {

            this.tm = getTaskManager();

            tasks = this.tm.getAllTasks  (username            ,
                                          credentials         ,
                                          this.exchangeFolder ) ;


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error reading tasks: " +
                                         e.getMessage(), e);
        }

        return getSyncItems(tasks, state);

    }

    /**
     * Create SyncItem from Task
     *
     * @param note
     * @return SyncItem
     * @trows SyncSourceException
     */
    private SyncItem getSyncItem(Task task)
    throws SyncSourceException {

        SyncItem syncItem = null;

        syncItem = new SyncItemImpl(this         ,
                                    task.getId() ,
                                    task.getState());

        try {

            if (this.isEncode()) {
                syncItem.setProperty(
                    new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                     Base64.encode(task.toPP().getBytes())
                ));
            } else {
                syncItem.setProperty(
                    new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                     task.toPP().getBytes())
                );
            }

        } catch (TaskParseException e) {
            throw new SyncSourceException( "Error setting the item "
                                          + syncItem
                                          , e
                                          );
        }

        return syncItem;

    }

    /**
     * Create SyncItem array from Task array
     *
     * @param task
     * @param state
     * @return SyncItem array
     * @trows SyncSourceException
     */
    private SyncItem[] getSyncItems(Task[] tasks,
                                    char   state)
    throws SyncSourceException {

        SyncItem[] syncItems = null;

        int l = tasks.length;

        syncItems = new SyncItem [l];

        for (int i=0; i < l; i++) {
            tasks    [i].setState(state)         ;
            syncItems[i] = getSyncItem(tasks[i]) ;
        }

        return syncItems;

    }

    /**
     *
     * Create TaskManager istance
     * (if not already exist)
     *
     * @trows DataAccessException
     */
    private TaskManager getTaskManager()
    throws DataAccessException {

        if (this.tm == null) {
            this.tm = new TaskManager(this.getHost() ,
                                      this.getPort());
        }

        return this.tm;

    }

}