/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.items.calendar.dao;

import java.io.IOException;
import java.sql.Timestamp;
import java.util.Date;
import java.util.ArrayList;
import java.text.MessageFormat;

import sync4j.exchange.httptransport.WebDavHttpTransport;
import sync4j.exchange.items.calendar.model.Calendar;
import sync4j.exchange.items.common.dao.ItemDAO;
import sync4j.exchange.AuthenticationException;
import sync4j.exchange.ExchangeAccessException;
import sync4j.exchange.xml.XmlParseException;
import sync4j.exchange.xml.XmlParser;
import sync4j.exchange.DataAccessException;

import sync4j.framework.server.store.NotFoundException;

/*
 * This class implements methods to access calendar data
 * in exchange server datastore
 *
 * @author Fabio Maggi @ Funambol
 * @version $Id: CalendarDAO.java,v 1.11 2005/05/20 10:19:34 luigiafassina Exp $
 *
 **/
public class CalendarDAO extends ItemDAO {

    //---------------------------------------------------------------- Constants
    public static final String TAG_ALLDAYEVENT        = "alldayevent"       ;

    public static final String TAG_DATE_END           = "dtend"             ;
    public static final String TAG_DATE_START         = "dtstart"           ;
    public static final String TAG_TIMEZONE_ID        = "timezoneid"        ;
    public static final String TAG_DESCRIPTION        =
        "textdescription"                                                    ;
    public static final String TAG_LOCATION           = "location"          ;
    public static final String TAG_RESPONSE           = "a:response"        ;
    public static final String TAG_SEQUENCE           = "sequence"          ;
    public static final String TAG_SUBJECT            = "subject"           ;
    public static final String TAG_DURATION           = "duration"          ;
    public static final String TAG_BUSY_STATUS        = "busystatus"        ;

    public static final String TAG_REPLUID            = "repluid"           ;

    public static final String TAG_LAST_MODIFIED      = "getlastmodified"   ;


    private static final String WEBDAV_MSG_PROPPATCH_UPDATE_CALENDAR         =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                          +
    "<D:propertyupdate xmlns:D=\"DAV:\" "                                    +
    "xmlns:EX=\"http://schemas.microsoft.com/exchange/\" "                   +
    "xmlns:HM=\"urn:schemas:calendar:\" "                                    +
    "xmlns:HN=\"urn:schemas:httpmail:\" >\n"                                 +
    "<D:set>\n"                                                              +
    " <D:prop>\n"                                                            +
    "  <D:contentclass>urn:content-classes:appointment</D:contentclass>\n"   +
    "  <EX:outlookmessageclass>IPM.Appointment</EX:outlookmessageclass>\n"   +
    "  <HN:subject>{0}</HN:subject>\n"                                       +
    "  <HN:textdescription>{1}</HN:textdescription>\n"                       +
    "  <HM:location>{2}</HM:location>\n"                                     +
    "  <HM:timezoneid>{3}</HM:timezoneid>\n"                                 +
    "  <HM:dtstart>{4}</HM:dtstart>\n"                                       +
    "  <HM:dtend>{5}</HM:dtend>\n"                                           +
    "  <HM:alldayevent>{6}</HM:alldayevent>\n"                               +
    "  <HM:duration>{7}</HM:duration>\n"                                     +
    "  <HM:busystatus>{8}</HM:busystatus>\n"                                 +
    "  <HM:sequence>{9}</HM:sequence>\n"                                     +
    " </D:prop>\n"                                                           +
    "</D:set>\n"                                                             +
    "</D:propertyupdate>"                                                    ;

    private static final String WEBDAV_MSG_SELECT_CALENDARS                  =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                          +
    "<D:searchrequest xmlns:D =\"DAV:\">\n"                                  +
    "<D:sql>\n"                                                              +
    "Select "                                                                +
    "\"http://schemas.microsoft.com/repl/repl-uid\" AS repluid, "            +
    "\"urn:schemas:calendar:location\" AS location, "                        +
    "\"urn:schemas:httpmail:subject\" AS subject, "                          +
    "\"urn:schemas:httpmail:textdescription\" AS textdescription, "          +
    "\"urn:schemas:calendar:timezoneid\" AS timezoneid, "                    +
    "\"urn:schemas:calendar:dtstart\" AS dtstart, "                          +
    "\"urn:schemas:calendar:dtend\" AS dtend, "                              +
    "\"urn:schemas:calendar:alldayevent\" AS alldayevent, "                  +
    "\"urn:schemas:calendar:sequence\" AS sequence, "                        +
    "\"urn:schemas:calendar:duration\" AS duration, "                        +
    "\"urn:schemas:calendar:busystatus\" AS busystatus, "                    +
    "\"DAV:getlastmodified\" AS getlastmodified, "                           +
    "\"DAV:isfolder\" AS isfolder "                                          +
    "FROM \"/{0}/{1}/{2}\"  "                                                +
    "{3}"                                                                    +
    "</D:sql>\n"                                                             +
    "</D:searchrequest>"                                                     ;

    private static final String FIELD_PREFIX          = "urn:schemas:calendar:";

    //------------------------------------------------------------- Private data


    private WebDavHttpTransport  webDavHttp          = null ;

    //------------------------------------------------------------- Costructors

    public CalendarDAO(String  exchangeServerHost,
                       int     exchangeServerPort)
        throws DataAccessException {

        this.webDavHttp         = new WebDavHttpTransport(exchangeServerHost,
                                                          exchangeServerPort);
    }
    //------------------------------------------------------------- Public methods

    /**
     * insert / update calendar<br>
     *
     * @param c
     * @param username
     * @param credentials
     * @param exchangeFolder
     *
     * @return new / updated calendar
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Calendar setCalendar(Calendar c              ,
                                String   username       ,
                                String   credentials    ,
                                String   exchangeFolder )
        throws DataAccessException {

        String     response                 = null ;

        String     id                       = null ;

        String     webDavCalendarMsg        = null ;
        String     webDavHeaderMsg          = null ;

        Date       date                     = null ;
        String     startDate                = null ;
        String     endDate                  = null ;

        String     timezoneId               = null ;

        String     server                   = null ;
        String     resource                 = null ;

        server   = getServerFromExchangeFolder   (exchangeFolder );
        resource = getResourceFromExchangeFolder (exchangeFolder );

        XmlParser  p = new XmlParser();

        timezoneId = String.valueOf(c.getTimezoneId());

        date = c.getDateStart();
        startDate = dateToWebDavTag(date);

        date = c.getDateEnd();
        endDate = dateToWebDavTag(date);

        webDavCalendarMsg
            = MessageFormat.format(WEBDAV_MSG_PROPPATCH_UPDATE_CALENDAR,
              new Object[] {
                p.getString    (c.getSubject         () ) ,
                p.getString    (c.getTextDescription () ) ,
                p.getString    (c.getLocation        () ) ,
                timezoneId                                ,
                startDate                                 ,
                endDate                                   ,
                String.valueOf (c.getAllDayEvent     () ) ,
                String.valueOf (c.getSequence        () ) ,
                p.getString    (c.getBusyStatus      () ) ,
                String.valueOf (c.getDuration        () )
              }
        );


        webDavHeaderMsg
            = MessageFormat.format(ItemDAO.WEBDAV_HEADER_PROPPATCH,
                new Object[] {
                        "/"                 +
                        server              +
                        "/"                 +
                        username            +
                        "/"                 +
                        resource                 ,
                        c.getHref()
              }
        );


        try {
             response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavCalendarMsg, FILE_ENCODING);
        } catch (Exception e) {
            throw new DataAccessException("Error parsing calendar", e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                   +
                                          username                  +
                                          " URI "                   +
                                          exchangeFolder            +
                                          " set Exchange calendar"  ,
                                          e)                        ;

        }

        try {
            id = XmlParser.getRuidFromResponse(response);
        } catch (XmlParseException e) {
            throw new DataAccessException("Error getting resource id", e);
        }

        c.setId(id);

        return c;

    }

    /**
     * Delete a <i>Calendar</i> from Exchange server.
     *
     * @param calendar
     * @param username
     * @param credentials
     * @param exchangeFolder
     *
     * @throws sync4j.exchange.util.DataAccessException
     *
     */
    public void removeCalendar(Calendar calendar       ,
                               String   username       ,
                               String   credentials    ,
                               String   exchangeFolder )
        throws DataAccessException {

        String webDavCalendarMsg       = null ;

        String webDavHeaderMsg         = null ;

        String response                = null ;

        String     server              = null ;
        String     resource            = null ;

        server   = getServerFromExchangeFolder   (exchangeFolder );
        resource = getResourceFromExchangeFolder (exchangeFolder );

        webDavCalendarMsg       = ""                      ;


        webDavHeaderMsg =
            MessageFormat.format(ItemDAO.WEBDAV_HEADER_REMOVE,
            new Object[] {
                            "/"                      +
                            server                   +
                            "/"                      +
                            username                 +
                            "/"                      +
                            resource                   ,
                            calendar.getHref()

            }
        );


        try {
             response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavCalendarMsg, FILE_ENCODING);
         } catch (Exception e) {
            throw new DataAccessException("Error parsing calendar", e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                        +
                                          username                       +
                                          " URI "                        +
                                          exchangeFolder                 +
                                          " removing Exchange calendar"  ,
                                          e)                             ;
        }

    }

    /**
     * get calendar array from Exchange Server
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     *
     * @return array of find calendars
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Calendar[] getCalendars(String     username       ,
                                   String     credentials    ,
                                   String[]   ids            ,
                                   String     exchangeFolder )
        throws DataAccessException {

        String clause = null ;

        clause = getClause(ids);

        return getCalendars(username, credentials, clause, exchangeFolder);
    }

    /**
     * get calendar array from Exchange Server
     *
     * @param username
     * @param credentials
     * @param fields
     * @param values
     * @param exchangeFolder
     * @return array of find calendars
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Calendar[] getCalendars(String username,
                                   String credentials,
                                   String fields[],
                                   Object values[],
                                   String exchangeFolder) throws DataAccessException {

        String clause                 = null ;

        clause = getClause(fields, values);

        return getCalendars(username, credentials, clause, exchangeFolder);
    }

    /**
     * Create a <i>Calendar</i> object from WebDav msg.
     *
     * @param id
     * @param webDavMsg
     *
     * @return a newly create <i>Calendar</i> object initialized with the
     *         fields in the result set
     *
     */
    private Calendar[] getCalendarsFromWebDavMsg(String webDavMsg)
    throws DataAccessException {

        ArrayList   calendars  = null ;
        String  []  resps      = null ;
        String  []  msg        = null ;

        String      response   = null ;
        String      id         = null ;
        String      replUid    = null ;

        String      isFolder   = null ;

        int         count      = 0    ;

        msg = new String[] {webDavMsg};

        calendars = new ArrayList();

        try {

            resps = XmlParser.getXMLTag(msg, TAG_RESPONSE);

            for (int i = 0, l = resps.length; i < l; i++) {

                response =  resps[i]                                         ;
                replUid  =  XmlParser.getXMLTagValue (response, TAG_REPLUID) ;
                id       =  getIdFromReplUid (replUid)                       ;

                isFolder =  XmlParser.getXMLInitTagValue (response,
                                                          TAG_IS_FOLDER)     ;

                if(PROP_NO_FOLDER.equals(isFolder)) {
                    calendars.add(getCalendarFromResponseTag(id, response));
                    count ++;
                }

            }

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing calendar",
                                          e);
        }

        return (Calendar[])calendars.toArray(new Calendar[count]);

    }

    /**
     * Create a <i>Calendar</i> object from WebDav msg.
     *
     * @param id
     * @param msgResponse
     *
     * @return a newly create <i>Calendar</i> object initialized with the
     *         fields in the result set
     *
     */
    private Calendar getCalendarFromResponseTag(String id, String msgResponse)
    throws DataAccessException {

        Calendar  c           = null ;
        String    lastUpdate  = null ;

        String    dateStart   = null ;
        String    dateEnd     = null ;

        String    value       = null ;

        c = new Calendar(id);

        try {

            c.setTextDescription (XmlParser.getXMLTagValue
                                    (msgResponse, TAG_DESCRIPTION)   ) ;

            c.setLocation        (XmlParser.getXMLTagValue
                                    (msgResponse, TAG_LOCATION)      ) ;

            c.setSubject         (XmlParser.getXMLTagValue
                                    (msgResponse, TAG_SUBJECT)       ) ;

            c.setBusyStatus      (XmlParser.getXMLTagValue
                                    (msgResponse, TAG_BUSY_STATUS)   ) ;

            dateStart  =         (XmlParser.getXMLInitTagValue
                                    (msgResponse, TAG_DATE_START)    ) ;

            dateEnd    =         (XmlParser.getXMLInitTagValue
                                    (msgResponse, TAG_DATE_END)      ) ;

            lastUpdate =         (XmlParser.getXMLInitTagValue
                                    (msgResponse, TAG_LAST_MODIFIED) ) ;

            value      =         (XmlParser.getXMLInitTagValue
                                    (msgResponse, TAG_ALLDAYEVENT)   ) ;

            if (value != null && value.length() > 0) {
                c.setAllDayEvent (Integer.parseInt(value));
            } else {
                c.setAllDayEvent (0);
            }

            value      =         (XmlParser.getXMLInitTagValue
                                                (msgResponse, TAG_DURATION)) ;

            if (value != null && value.length() > 0) {
                c.setDuration (Long.parseLong(value));
            } else {
                c.setDuration (0);
            }

            value = (XmlParser.getXMLInitTagValue
                        (msgResponse, TAG_SEQUENCE));

            if (value != null && value.length() > 0) {
                c.setSequence (Integer.parseInt(value));
            } else {
                c.setSequence (0);
            }

            value = (XmlParser.getXMLInitTagValue
                        (msgResponse, TAG_TIMEZONE_ID));

            if (value != null && value.length() > 0) {
                c.setTimezoneId (Integer.parseInt(value));
            } else {
                c.setTimezoneId (0);
            }

            c.setDateStart (XmlParser.webDavTagToDate(dateStart) ) ;
            c.setDateEnd   (XmlParser.webDavTagToDate(dateEnd)   ) ;

            c.setLastModified (XmlParser.webDavTagToDate(lastUpdate));

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing calendar item: ", e);
        }

        return c;

    }

    /**
     * Build a clause
     * to webdav select request
     * between ids item array
     *
     * @param ids
     * @return webdav select clause
     **/
    private static String getClause (String[] fields, Object[] values) {

        StringBuffer clause = new StringBuffer();

        int l = fields.length;

        if (l == 0) {
            return "";
        }

        clause.append("where ");

        StringBuffer tmp = null;
        String valueTmp  = null;

        for (int i = 0; i < l; i++) {
            tmp = new StringBuffer();

            valueTmp = "";

            if (TAG_DATE_START.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("dtstart\"=CAST(\"");

                if (values[i] != null) {
                    valueTmp = dateToWebDavTag((java.util.Date)values[i]);
                }
                tmp.append(valueTmp).append("\" as 'dateTime')");
            }

            if (TAG_DATE_END.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("dtend\"=CAST(\"");

                if (values[i] != null) {
                    valueTmp = dateToWebDavTag((java.util.Date)values[i]);
                }

                tmp.append(valueTmp).append("\" as 'dateTime')");
            }

            if (TAG_LOCATION.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("location\"='");

                if (values[i] != null) {
                    valueTmp = String.valueOf(values[i]);
                }
                tmp.append(valueTmp).append("'");
            }

            if (i != 0 && tmp.length() != 0) {
                clause.append(" and ").append(tmp);
            } else if (tmp.length() != 0) {
                clause.append(tmp);
            }

        }

        clause.append("\r\n");

        return clause.toString();

    }


    /**
     * get calendars from Exchange Server with the given clause
     *
     * @param username
     * @param credentials
     * @param clause
     * @param exchangeFolder
     * @return array of find calendars
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    private Calendar[] getCalendars(String username,
                                    String credentials,
                                    String clause,
                                    String exchangeFolder) throws DataAccessException {

        Calendar[] exchangeCalendars = null;

        String response = null;
        String webDavCalendarMsg = null;

        String webDavHeaderMsg = null;

        String server = null;
        String resource = null;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        webDavCalendarMsg =
                MessageFormat.format(WEBDAV_MSG_SELECT_CALENDARS,
                                     new Object[] {server, username, resource,
                                     clause});

        webDavHeaderMsg =
                MessageFormat.format(ItemDAO.WEBDAV_HEADER_SELECT,
                                     new Object[] {
                                         "/" +
                                         server +
                                         "/" +
                                         username +
                                         "/" +
                                         resource
                                     }
                );

        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials,
                                                   webDavCalendarMsg, FILE_ENCODING);
        } catch (IOException e) {
            throw new DataAccessException(
                    "Error getting Exchange server response",
                    e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                                          username +
                                          " URI " +
                                          exchangeFolder +
                                          " getting exchange calendar",
                                          e);

        }

        exchangeCalendars = getCalendarsFromWebDavMsg(response);

        return exchangeCalendars;

    }
}
