/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.items.common.dao;

import java.io.IOException;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Timestamp;

import java.text.MessageFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Date;

import sync4j.exchange.httptransport.WebDavHttpTransport;
import sync4j.exchange.items.common.model.Item;
import sync4j.exchange.AuthenticationException;
import sync4j.exchange.ExchangeAccessException;
import sync4j.exchange.xml.XmlParseException;
import sync4j.exchange.xml.XmlParser;

import sync4j.exchange.util.DataAccess;
import sync4j.exchange.DataAccessException;

import sync4j.framework.server.store.NotFoundException;

/*
 * This class implement methods to access items common data
 * in database source and exchange datastore
 *
 * @author Fabio Maggi @ Funambol
 * @version $Id: ItemDAO.java,v 1.11 2005/05/11 15:59:35 nichele Exp $
 *
 **/
public class ItemDAO {

    //--------------------------------------------------------------------- Constants

    protected static final String FILE_ENCODING = "UTF-8";


    protected static final String WEBDAV_HEADER_PROPPATCH                  =
    "PROPPATCH {0}/{1} HTTP/1.0 \r\n";

    protected static final String WEBDAV_HEADER_REMOVE                     =
    "DELETE {0}/{1} HTTP/1.0\r\n";

    protected static final String WEBDAV_HEADER_SELECT                     =
    "SEARCH {0} HTTP/1.0\r\n"                                              +
    "Depth: 1,noroot\r\n";

    private static final String WEBDAV_HEADER_PROPFIND                     =
    "PROPFIND {0} HTTP/1.0\r\n"                                            +
    "Depth: 1,noroot\r\n";

    private static final String WEBDAV_MSG_SELECT_ITEM                     =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                        +
    "<D:searchrequest xmlns:D =\"DAV:\">\n"                                +
    "<D:sql>\n"                                                            +
    "Select "                                                              +
    "\"http://schemas.microsoft.com/repl/repl-uid\" AS repluid, "          +
    "\"DAV:getlastmodified\" AS getlastmodified, "                         +
    "\"DAV:creationdate\" AS creationdate, "                               +
    "\"DAV:href\" AS href, "                                               +
    "\"DAV:isfolder\" AS isfolder "                                        +
    "FROM \"/{0}/{1}/{2}\"  "                                              +
    "</D:sql>\n"                                                           +
    "</D:searchrequest>"                                                   ;

    private static final String WEBDAV_MSG_PROPFIND_SELECT_ITEM            =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                        +
    "<D:propfind "                                                         +
    "xmlns:D=\"DAV:\" "                                                    +
    "xmlns:EN=\"http://schemas.microsoft.com/repl/\""                      +
    ">\n"                                                                  +
    "<D:prop><D:href/></D:prop>\n"                                         +
    "<D:prop><D:creationdate/></D:prop>\n"                                 +
    "<D:prop><D:getlastmodified/></D:prop>\n"                              +
    "<D:prop><EN:repl-uid/></D:prop>\n"                                    +
    "</D:propfind>\n"                                                      ;

    private static final String CLAUSE_REPLID_FIELD
        = "\"http://schemas.microsoft.com/repl/repl-uid\""          ;

    private static final String CONTENT_REPLID_START   = "rid:"     ;


    private static final String SQL_SELECT_ITEMS_BY_PRINCIPAL_BY_SOURCE    =
    "select id, last_modified "                                            +
    "from ss_exch_item_status "                                            +
    "where "                                                               +
    "uri = ? and principal = ?"                                     ;

    private static final String SQL_ADD_LOCAL_ITEM                         =
    "insert into ss_exch_item_status "                                     +
    "(id, last_modified, uri, principal) "                                 +
    "values(?, ?, ?, ?)"                                                   ;

    private static final String SQL_DELETE_LOCAL_ITEMS                     =
    "delete from ss_exch_item_status "                                     +
    "where uri = ? and principal = ? "                                     ;

    private static final String TAG_HREF             = "href"              ;
    private static final String TAG_CREATION_DATE    = "creationdate"      ;
    private static final String TAG_REPLUID          = "repluid"           ;
    private static final String TAG_LAST_MODIFIED    = "getlastmodified"   ;
    protected
            static final String TAG_IS_FOLDER        = "isfolder"          ;
    private static final String TAG_PROP             = "a:prop"            ;

    protected
        static final String PROP_NO_FOLDER           = "0"                 ;

    protected static final String NEW_LINE           = "\r\n"              ;

    private static final String
        STATUS_ACCESS_DANIED                         = "401"               ;
    private static final String
        STATUS_OK_LOAD_CHANGE                        = "207"               ;
    private static final String
        STATUS_OK_DELETE                             = "200"               ;
    private static final String
        STATUS_NOT_FOUND                             = "404"               ;

    private static final String DATE_FORMAT_WEBDAV   =
                   "yyyy-MM-dd HH:mm:ss"                                   ;
    //--------------------------------------------------------------------- Private data

    private DataAccess           da                  = null ;
    private WebDavHttpTransport  webDavHttp          = null ;

    //--------------------------------------------------------------------- Costructors

    public ItemDAO ()
        throws DataAccessException {

        this.da = new DataAccess ()  ;

    }

    public ItemDAO (String  host ,
                    int     port )
    throws DataAccessException {

        this.da          = new DataAccess ()  ;
        this.webDavHttp  = new WebDavHttpTransport(host, port);

    }

    //--------------------------------------------------------------------- Public methods

    /**
     * get local items
     *
     * @param sourceURI
     * @param principal
     *
     * @return array of find local items
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Item[] getLocalItems (String sourceURI, String principal)
    throws DataAccessException {

         Connection         connection     = null ;
         PreparedStatement  ps             = null ;
         ResultSet          rs             = null ;

         ArrayList          rows           = null ;

         Item[]             items          = null ;
         Item               item           = null ;

         int                count          = 0    ;

         try {

             connection = this.da.getConnection();

             rows = new ArrayList();

             ps = connection.prepareStatement(
                    SQL_SELECT_ITEMS_BY_PRINCIPAL_BY_SOURCE);

             ps.setString (1, sourceURI ) ;
             ps.setString (2, principal ) ;

             rs = ps.executeQuery();

             while(rs.next()) {
                 rows.add(new Item(rs.getString(1),
                          new Date(rs.getTimestamp(2).getTime())));
                 ++count;
             }

             items = (Item[]) rows.toArray(new Item[count]);

         } catch (Exception e) {
             throw new DataAccessException("Error getting local items", e);
         } finally {
             try {
                 this.da.cleanUp(connection, ps, rs);
             } catch (Exception e)  {
                 throw new DataAccessException("Error getting local items", e);
             }
         }

         return items;

    }

    /**
     * delete local items
     *
     * @param items
     * @param sourceURI
     * @param principal
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public void addLocalItems(Item[] items     ,
                              String sourceURI ,
                              String principal
                              )
    throws DataAccessException {

         Connection         connection    = null ;
         PreparedStatement  ps            = null ;

         String             id            = null;
         Date               lastModified  = null;

         try {

            connection = this.da.getConnection();

            for (int i=0, l = items.length; i < l; i++) {

                ps = connection.prepareStatement(SQL_ADD_LOCAL_ITEM);

                id            = items[i].getId()           ;
                lastModified  = items[i].getLastModified() ;

                ps.setString    (1, id                                    ) ;
                ps.setTimestamp (2, new Timestamp(lastModified.getTime()) ) ;
                ps.setString    (3, sourceURI                             ) ;
                ps.setString    (4, principal                             ) ;

                ps.executeUpdate();

                ps.close() ;
                ps = null  ;

            }


         } catch (Exception e) {
             throw new DataAccessException("Error adding local items", e);
         } finally {
             try {
                 this.da.cleanUp(connection, ps, null);
             } catch (Exception e)  {
                 throw new DataAccessException("Error adding local items", e);
             }
         }

    }

    /**
     * delete local items
     *
     * @param sourceURI
     * @param principal
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public void deleteLocalItems(String sourceURI, String principal)
    throws DataAccessException {

         Connection         connection  = null ;
         PreparedStatement  ps          = null ;

         try {

            connection = this.da.getConnection();

            ps = connection.prepareStatement(SQL_DELETE_LOCAL_ITEMS);

            ps.setString  (1, sourceURI ) ;
            ps.setString  (2, principal ) ;

            ps.executeUpdate();

         } catch (Exception e) {
             throw new DataAccessException("Error deleting local items", e);
         } finally {
             try {
                 this.da.cleanUp(connection, ps, null);
             } catch (Exception e)  {
                 throw new DataAccessException("Error deleting local items", e);
             }
         }

    }

    /**
     * get all sourceURI type items from Exchange Server
     *
     * @param username
     * @param credentials
     * @param exchangeFolder
     *
     * @return array of find items
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Item[] getAllExchangeItems(String username       ,
                                      String credentials    ,
                                      String exchangeFolder )
    throws DataAccessException {

        ArrayList items                = null ;
        Item      item                 = null ;

        String[]  props                = null ;
        String    prop                 = null ;
        String[]  resp                 = null ;

        String    replUid              = null ;
        String    uri                  = null ;
        String    creationDate         = null ;
        String    lastUpdate           = null ;

        String    isFolder             = null ;

        String    response             = null ;

        String    webDavBodyMsg        = null ;

        String    webDavHeaderMsg      = null ;

        String    server               = null ;
        String    resource             = null ;

        int       count                = 0    ;

        server   = getServerFromExchangeFolder   (exchangeFolder );
        resource = getResourceFromExchangeFolder (exchangeFolder );

        webDavBodyMsg =
                    MessageFormat.format(WEBDAV_MSG_SELECT_ITEM,
            new Object[] {server, username, resource});

        webDavHeaderMsg =
            MessageFormat.format(WEBDAV_HEADER_SELECT,

            new Object[] {
                    "/"                  +
                        server           +
                        "/"              +
                        username         +
                        "/"              +
                        resource
            }
        );


        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavBodyMsg, FILE_ENCODING);
        } catch (IOException e) {
            throw new DataAccessException("Error getting Exchange server response",
                                           e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                    +
                                          username                   +
                                          " URI "                    +
                                          exchangeFolder             +
                                          " reading exchange items"  ,
                                          e)                         ;
        }

        resp = new String[] {response};

        try {
            props = XmlParser.getXMLTag(resp, TAG_PROP);
        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing item", e);
        }

        items = new ArrayList();

        for (int i=0, l = props.length; i < l; i++) {

            try {

                prop = props[i];

                isFolder      = XmlParser.getXMLInitTagValue
                                 (prop, TAG_IS_FOLDER)        ;

                replUid       = XmlParser.getXMLTagValue
                                 (prop, TAG_REPLUID)          ;

                uri           = XmlParser.getXMLTagValue
                                 (prop, TAG_HREF)             ;

                creationDate  = XmlParser.getXMLInitTagValue
                                 (prop, TAG_CREATION_DATE)    ;

                lastUpdate    = XmlParser.getXMLInitTagValue
                                 (prop, TAG_LAST_MODIFIED)    ;

                item = new Item(getIdFromReplUid(replUid)               ,
                                XmlParser.getHrefFromUri(uri)           ,
                                XmlParser.webDavTagToDate(creationDate) ,
                                XmlParser.webDavTagToDate(lastUpdate));

                if(PROP_NO_FOLDER.equals(isFolder)) {
                    items.add(item);
                    count ++;
                }

            } catch (XmlParseException e) {
                throw new DataAccessException("Error parsing item", e);
            }

        }

        return (Item[])items.toArray(new Item[count]);
    }

    /**
     * Check response message from Exchange server
     *
     * @param msg the message to be checked
     *
     * @throws AuthenticationException in case the request has not been
     *                                 authenticated by the server
     * @throws ExchangeAccessException in case of other errors
     *
     **/
    protected void checkResponseStatus(String msg)
    throws AuthenticationException ,
           ExchangeAccessException ,
           NotFoundException
    {

        String header = null;

        if (msg == null && msg.length() < 30) {
            throw new ExchangeAccessException
                ("Exchange server, wrong response: " + msg);
        }

        header = msg.substring(0, 30);

        if (header.indexOf(STATUS_ACCESS_DANIED) != -1) {
            throw new AuthenticationException
                ("Exchange server, access denied");
        }

        if (header.indexOf(STATUS_OK_LOAD_CHANGE ) == -1 &&
            header.indexOf(STATUS_OK_DELETE      ) == -1 &&
            header.indexOf(STATUS_NOT_FOUND      ) == -1 ) {
            throw new ExchangeAccessException
                ("Exchange server, error accessing resource");
        }

        if (header.indexOf(STATUS_NOT_FOUND     ) != -1) {
            throw new NotFoundException
                ("Exchange server, resource not found");
        }

    }

    /**
     * Build a clause
     * to webdav select request
     * between ids item array
     *
     * @param ids
     * @return webdav select clause
     **/
    public static String getClause (String[] ids) {

        String clause = null;

        int l = ids.length;

        if (l == 0) {
            return "";
        }

        clause = "where "             +
                 CLAUSE_REPLID_FIELD  +
                 " = '"               +
                 CONTENT_REPLID_START +
                 ids[0]               +
                 "'"                  ;

        if (l > 1) {

            for (int i = 1; i < l; i++) {
                clause = clause       +
                " OR "                +
                CLAUSE_REPLID_FIELD   +
                " = '"                +
                CONTENT_REPLID_START  +
                ids[i]                +
                "'"                   ;
            }

        }

        clause = clause + "\r\n";

        return clause;

    }

    /**
     * Make a id from replid webdavtag.
     *
     * @param replUid
     * @return id
     * @throws XmlParseException
     **/
    public static String getIdFromReplUid(String replUid)
    throws XmlParseException {

        int start = 0;

        start = replUid.indexOf(CONTENT_REPLID_START);

        if (start == -1) {
            throw new XmlParseException("Error in replUid-TAG, value: " + replUid);
        }

        return replUid.substring(start + CONTENT_REPLID_START.length());
    }

    /**
     * Make a webdav tag date from a date
     *
     * @param date date
     * @return webdav tag date
     * @throws XmlParseException
     **/
    public static String dateToWebDavTag(Date date) {

        SimpleDateFormat formatter = null;
        String value = null;
        String dateTag = null;

        if (date != null) {
            formatter = new SimpleDateFormat(DATE_FORMAT_WEBDAV);

            value = formatter.format(date);

            dateTag = value.substring(0, 10) +
                      "T" +
                      value.substring(11) +
                      ".000Z";
        } else {
            dateTag = "";
        }

        return dateTag;
    }

    /**
     * Make a webdav tag date from a date
     *
     * @param date webdav tag content
     * @param timezone
     * @return webdav tag date
     * @throws XmlParseException
     **/
    public static String dateToWebDavTag (Date date, String timezone)
    throws XmlParseException {

        SimpleDateFormat  formatter    = null ;
        String            value        = null ;
        String            dateTag      = null;

        try {

            if (date != null) {
                formatter = new SimpleDateFormat (DATE_FORMAT_WEBDAV);

                value = formatter.format(date);

                dateTag =  value.substring(0, 10) +
                           "T"                    +
                           value.substring(11)    +
                           ".000Z"                ;
            } else {
                dateTag = "";
            }

        } catch (Exception e) {
            throw new XmlParseException(e.toString());
        }

        return dateTag;
    }

    //--------------------------------------------------------------------- Protected methods

    protected String getServerFromExchangeFolder(String uri) {
        return uri.substring(0, uri.indexOf("/"));
    }

    protected String getResourceFromExchangeFolder(String uri) {
        return uri.substring(uri.indexOf("/")  + 1);
    }

}
