/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.items.task.dao;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;

import sync4j.exchange.httptransport.WebDavHttpTransport;
import sync4j.exchange.items.task.model.Task;
import sync4j.exchange.items.common.dao.ItemDAO;
import sync4j.exchange.DataAccessException;
import sync4j.exchange.xml.XmlParseException;
import sync4j.exchange.xml.XmlParser;


/*
 * This class implements methods to access task data
 * in exchange server datastore
 *
 * @author Fabio Maggi @ Funambol
 * @version $Id: TaskDAO.java,v 1.9 2005/05/11 16:01:54 nichele Exp $
 *
 **/
public class TaskDAO extends ItemDAO {

    //---------------------------------------------------------------- Constants

    private static final String WEBDAV_MSG_PROPPATCH_UPDATE_TASK            =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                         +
    "<D:propertyupdate xmlns:D=\"DAV:\" "                                   +
    "xmlns:EX=\"http://schemas.microsoft.com/exchange/\" "                  +
    "xmlns:HN=\"urn:schemas:httpmail:\">\n"                                 +
    "<D:set>\n"                                                             +
    " <D:prop>\n"                                                           +
    "  <D:contentclass>urn:content-classes:person</D:contentclass>\n"       +
    "  <EX:outlookmessageclass>IPM.Task</EX:outlookmessageclass>\n"         +
    "  <HN:subject>{0}</HN:subject>\n"                                      +
    "  <HN:subject>{1}</HN:subject>\n"                                      +
    "  <HN:textdescription>{2}</HN:textdescription>\n"                      +
    "  <EX:date>{3}</EX:date>\n"                                            +
    " </D:prop>\n"                                                          +
    "</D:set>\n"                                                            +
    "</D:propertyupdate>"                                                   ;

    private static final String WEBDAV_MSG_SELECT_TASKS                     =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                         +
    "<D:searchrequest xmlns:D =\"DAV:\">\n"                                 +
    "<D:sql>\n"                                                             +
    "Select "                                                               +
    "\"urn:schemas:httpmail:textdescription\" AS textdescription, "         +
    "\"urn:schemas:httpmail:subject\" AS subject, "                         +
    "\"urn:schemas:httpmail:date\" AS date, "                               +
    "\"urn:schemas:calendar:dtstart\" AS dtstart, "                         +
    "\"urn:schemas:calendar:dtend\" AS dtend, "                             +
    "\"http://schemas.microsoft.com/repl/repl-uid\" AS repluid, "           +
    "\"DAV:getlastmodified\" AS getlastmodified, "                          +
     "\"DAV:isfolder\" AS isfolder "                                        +
    "FROM \"/{0}/{1}/{2}\"  "                                               +
    "{3} "                                                                  +
    "</D:sql>\n"                                                            +
    "</D:searchrequest>"                                                    ;

    private static final String SERVER_PATH            = "/Exchange/"       ;

    private static final String TAG_DATE               = "date"             ;
    private static final String TAG_DESCRIPTION        =
        "textdescription"                                                   ;
    private static final String TAG_RESPONSE           = "a:response"       ;
    private static final String TAG_SUBJECT            = "subject"          ;

    private static final String TAG_REPLUID            = "repluid"          ;

    private static final String TAG_LAST_MODIFIED     = "getlastmodified"   ;

    private static final String NEW_LINE               = "\r\n"             ;

    //------------------------------------------------------------- Private data


    private WebDavHttpTransport  webDavHttp          = null ;

    //------------------------------------------------------------- Costructors

    public TaskDAO(String exchangeServerHost ,
                   int    exchangeServerPort)
        throws DataAccessException {

        this.webDavHttp         = new WebDavHttpTransport(exchangeServerHost,
                                                          exchangeServerPort);
    }
    //------------------------------------------------------------- Public methods

    /**
     * insert / update task
     *
     * @param task
     * @param username
     * @param credentials
     * @param exchangeFolder
     *
     * @return new / updated task
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Task setTask(Task    task           ,
                        String  username       ,
                        String  credentials    ,
                        String  exchangeFolder )
        throws DataAccessException {


        String response                = null ;

        String id                      = null ;

        String webDavTaskMsg           = null ;
        String webDavHeaderMsg         = null ;

        String date                    = null ;

        String server                  = null ;
        String resource                = null ;

        server   = getServerFromExchangeFolder   (exchangeFolder );
        resource = getResourceFromExchangeFolder (exchangeFolder );

        date = dateToWebDavTag(task.getDate());

        webDavTaskMsg
            = MessageFormat.format(WEBDAV_MSG_PROPPATCH_UPDATE_TASK,
              new Object[] {
                task.getSubject()          ,
                task.getSubject()          ,
                task.getTextDescription()  ,
                date
              }
        );

        webDavHeaderMsg
            = MessageFormat.format(ItemDAO.WEBDAV_HEADER_PROPPATCH,
                new Object[] {
                    "/"                   +
                        server            +
                        "/"               +
                        username          +
                        "/"               +
                        resource             ,
                    task.getHref()
              }
        );


        try {
             response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavTaskMsg, FILE_ENCODING);
        } catch (Exception e) {
            throw new DataAccessException("Error getting Exchange server response",
                                          e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                   +
                                          username                  +
                                          " URI "                   +
                                          exchangeFolder            +
                                          " setting exchange task"  ,
                                          e)                        ;

        }

        try {
            id = XmlParser.getRuidFromResponse(response);
        } catch (XmlParseException e) {
            throw new DataAccessException("Error getting task id", e);
        }

        task.setId(id);

        return task;

    }

    /**
     * Delete a <i>Task</i> from Exchange server.
     *
     * @param task
     * @param username
     * @param credentials
     * @param exchangeFolder
     *
     * @throws sync4j.exchange.util.DataAccessException
     *
     */
    public void removeTask(Task    task           ,
                           String  username       ,
                           String  credentials    ,
                           String  exchangeFolder )
        throws DataAccessException {

        String webDavTaskMsg           = null ;

        String webDavHeaderMsg         = null ;

        String response                = null ;

        String server                  = null ;
        String resource                = null ;

        server   = getServerFromExchangeFolder   (exchangeFolder );
        resource = getResourceFromExchangeFolder (exchangeFolder );

        webDavTaskMsg          = ""                ;

        webDavHeaderMsg =
            MessageFormat.format(ItemDAO.WEBDAV_HEADER_REMOVE,
            new Object[] {
                        "/"                     +
                            server              +
                            "/"                 +
                            username            +
                            "/"                 +
                            resource              ,
                        task.getHref()
            }
        );

        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavTaskMsg, FILE_ENCODING);
        } catch (IOException e) {
            throw new DataAccessException("Error getting Exchange server response",
                                          e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                     +
                                          username                    +
                                          " URI "                     +
                                          exchangeFolder              +
                                          " removing exchange task"   ,
                                          e)                          ;

        }

    }

    /**
     * get tasks from Exchange Server
     *
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     *
     * @return array of find tasks
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Task[] getTasks (String    username       ,
                            String    credentials    ,
                            String[]  ids            ,
                            String    exchangeFolder )
        throws DataAccessException {

        Task[] exchangeTasks          = null ;

        String response               = null ;
        String webDavTaskMsg          = null ;
        String webDavHeaderMsg        = null ;

        String clause                 = null ;

        String server                  = null ;
        String resource                = null ;

        server   = getServerFromExchangeFolder   (exchangeFolder );
        resource = getResourceFromExchangeFolder (exchangeFolder );

        clause = getClause(ids);

        webDavTaskMsg =
            MessageFormat.format(WEBDAV_MSG_SELECT_TASKS,
            new Object[] {server, username, resource, clause});

        webDavHeaderMsg =
            MessageFormat.format(ItemDAO.WEBDAV_HEADER_SELECT,

            new Object[] {
                        "/"                       +
                            server                +
                            "/"                   +
                            username              +
                            "/"                   +
                        resource
            }
        );


        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavTaskMsg, FILE_ENCODING);
        } catch (IOException e) {
            throw new DataAccessException("Error getting Exchange server response",
                                          e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                   +
                                          username                  +
                                          " URI "                   +
                                          exchangeFolder            +
                                          " getting exchange task"  ,
                                          e)                        ;

        }

        exchangeTasks = getTasksFromWebDavMsg(response);

        return exchangeTasks;

    }

    //------------------------------------------------------------- Private methods

    /**
     * Create a <i>Task</i> array object from WebDav msg.
     *
     * @param id
     * @param webDavMsg
     *
     * @return a newly create <i>Task</i> object initialized with the
     *         fields in the result set
     *
     */
    private Task[] getTasksFromWebDavMsg(String webDavMsg)
    throws DataAccessException {

        ArrayList  tasks      = null ;
        String []  resps      = null ;
        String []  msg        = null ;

        String     response   = null ;
        String     id         = null ;
        String     replUid    = null ;

        String     isFolder   = null ;
        int        count      = 0    ;

        msg = new String[] {webDavMsg};

        tasks = new ArrayList();

        try {

            resps = XmlParser.getXMLTag(msg, TAG_RESPONSE);

            for (int i = 0, l = resps.length; i < l; i++) {

                response =  resps[i]                                           ;
                replUid  =  XmlParser.getXMLInitTagValue(response, TAG_REPLUID);
                id       =  getIdFromReplUid(replUid)                          ;

                isFolder    =  XmlParser.getXMLInitTagValue(response,
                                                            TAG_IS_FOLDER)  ;
                if(PROP_NO_FOLDER.equals(isFolder)) {
                    tasks.add(getTaskFromResponseTag(id, response));
                    count ++;
                }

            }

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing task", e);
        }

        return (Task[])tasks.toArray(new Task[count]);

    }

    /**
     * Create a <i>Task</i> object from response tag.
     *
     * @param id
     * @param msgResponse
     *
     * @return a newly create <i>Task</i> object initialized with the
     *         fields in the result set
     *
     */
    private Task getTaskFromResponseTag(String id, String msgResponse)
    throws DataAccessException {

        Task    t           = null ;
        String  lastUpdate  = null ;
        String  date        = null ;

        t = new Task(id);

        try {

            t.setTextDescription         (XmlParser.getXMLInitTagValue
                                            (msgResponse, TAG_DESCRIPTION))   ;

            t.setSubject                 (XmlParser.getXMLInitTagValue
                                            (msgResponse, TAG_SUBJECT))       ;

            date       =                  XmlParser.getXMLInitTagValue
                                            (msgResponse, TAG_DATE)           ;

            lastUpdate =                  XmlParser.getXMLInitTagValue
                                            (msgResponse, TAG_LAST_MODIFIED)  ;

            t.setDate                    (XmlParser.webDavTagToDate(date))    ;

            t.setLastModified            (XmlParser.webDavTagToDate
                                            (lastUpdate))                     ;

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing task item: ", e);
        }

        return t;

    }

}
